//! Rocks DB implementation of the storage layer
use crate::{
    util::{load_events_and_filter, SecondaryIndex},
    Error, Storage,
};
use nostr_rs_types::types::{Event, Filter, Tag};
use rocksdb::{
    BoundColumnFamily, ColumnFamilyDescriptor, IteratorMode, Options, SliceTransform, WriteBatch,
    DB,
};
use std::{collections::HashSet, path::Path, sync::Arc};

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
enum ReferenceType {
    Events,
    Author,
    RefPublicKey,
    RefEvent,
    Kind,
    LocalEvents,
}

impl ReferenceType {
    #[inline]
    pub fn as_str(&self) -> &'static str {
        match self {
            Self::Events => "events",
            Self::Author => "authors",
            Self::RefPublicKey => "refs_by_public_key",
            Self::RefEvent => "refs_by_ids",
            Self::Kind => "kinds",
            Self::LocalEvents => "local",
        }
    }
}

/// Storage implementation using RocksDB
pub struct RocksDb {
    db: DB,
}

impl RocksDb {
    /// Creates a new instance passing rocksDb options and a path
    pub fn with_custom_db<T: AsRef<Path>>(options: Options, path: T) -> Result<Self, Error> {
        let db = DB::open_cf_descriptors(
            &options,
            path,
            vec![
                ColumnFamilyDescriptor::new(ReferenceType::Events.as_str(), options.clone()),
                ColumnFamilyDescriptor::new(ReferenceType::Author.as_str(), options.clone()),
                ColumnFamilyDescriptor::new(ReferenceType::RefEvent.as_str(), options.clone()),
                ColumnFamilyDescriptor::new(ReferenceType::RefPublicKey.as_str(), options.clone()),
                ColumnFamilyDescriptor::new(ReferenceType::Kind.as_str(), options.clone()),
                ColumnFamilyDescriptor::new(ReferenceType::LocalEvents.as_str(), options.clone()),
            ],
        )?;
        Ok(Self { db })
    }

    /// Creates a new instance, passing just the path, the default options are passed
    pub fn new<T: AsRef<Path>>(path: T) -> Result<Self, Error> {
        let mut options = Options::default();
        options.create_if_missing(true);
        options.create_missing_column_families(true);
        options.set_compression_type(rocksdb::DBCompressionType::Snappy);
        options.optimize_for_point_lookup(256 * 1024 * 1024);

        let prefix_extractor = SliceTransform::create_fixed_prefix(1);
        options.set_prefix_extractor(prefix_extractor);

        Self::with_custom_db(options, path)
    }

    #[inline]
    fn reference_to_cf_handle(
        &self,
        namespace: ReferenceType,
    ) -> Result<Arc<BoundColumnFamily>, Error> {
        self.db
            .cf_handle(namespace.as_str())
            .ok_or(Error::InvalidColumnFamily)
    }

    /// Instead of loading the whole event, only the event id is loaded. This is
    /// done for performance, to avoid loading the same event multiple times,
    /// also to only load from the database the events that are needed
    /// (according to limits and other constraints)
    fn get_event_ids_by_ns_and_prefix(
        &self,
        cf_handle: &Arc<BoundColumnFamily>,
        prefix: &[u8],
        limit: Option<usize>,
    ) -> Result<Vec<Vec<u8>>, Error> {
        let mut items = vec![];
        let limit = limit.unwrap_or(usize::MAX);

        for item in self.db.prefix_iterator_cf(cf_handle, prefix) {
            let (key, value) = match item {
                Ok((k, v)) => (k, v),
                Err(e) => return Err(Error::RocksDb(e)),
            };

            if !key.starts_with(prefix) {
                break;
            }

            items.push(value.to_vec());
            if items.len() >= limit {
                break;
            }
        }

        Ok(items)
    }

    /// Loads all events that references a given event. The reference-type is provided and it can be any value of `ReferenceType`.
    fn get_event_referenced_as<T: AsRef<[u8]>>(
        &self,
        cf_handle: &Arc<BoundColumnFamily>,
        id: T,
        limit: Option<usize>,
    ) -> Result<HashSet<Vec<u8>>, Error> {
        let mut items = HashSet::new();
        let limit = limit.unwrap_or(usize::MAX);
        let prefix = id.as_ref();

        for item in self.db.prefix_iterator_cf(cf_handle, prefix) {
            let (key, value) = match item {
                Ok((k, v)) => (k, v),
                Err(e) => return Err(Error::RocksDb(e)),
            };

            if !key.starts_with(prefix) {
                break;
            }

            items.insert(value.to_vec());
            if limit == items.len() {
                break;
            }
        }

        Ok(items)
    }
}

impl Storage for RocksDb {
    fn get_local_events<C>(&self, for_each: C) -> Result<(), Error>
    where
        C: Fn(Event) -> Result<(), Error>,
    {
        let cf_handle = self.reference_to_cf_handle(ReferenceType::LocalEvents)?;
        self.db
            .iterator_cf(&cf_handle, IteratorMode::Start)
            .into_iter()
            .map(|res| {
                if let Ok((key, _)) = res {
                    if let Ok(Some(event)) = self.get_event(&key) {
                        for_each(event)
                    } else {
                        Ok(())
                    }
                } else {
                    Ok(())
                }
            })
            .collect::<Result<Vec<_>, Error>>()?;
        Ok(())
    }

    fn set_local_event(&self, event: &Event) -> Result<(), Error> {
        self.db.put_cf(
            &self.reference_to_cf_handle(ReferenceType::LocalEvents)?,
            *(event.id),
            &[],
        )?;
        Ok(())
    }

    fn store(&self, event: &Event) -> Result<bool, Error> {
        let event_id = event.id.clone();

        if let Ok(Some(_)) = self.db.get_cf(
            &self.reference_to_cf_handle(ReferenceType::Events)?,
            *event_id,
        ) {
            return Ok(false);
        }
        let secondary_index = SecondaryIndex::new(&event_id, event.created_at());
        let author_id = secondary_index.new_id(event.author());
        let json = serde_json::to_vec(event)?;
        let kind: u32 = event.kind().into();
        let kind_id = secondary_index.new_id(&kind.to_be_bytes());

        let mut buffer = WriteBatch::default();

        buffer.put_cf(
            &self.reference_to_cf_handle(ReferenceType::Events)?,
            *event_id,
            json,
        );
        buffer.put_cf(
            &self.reference_to_cf_handle(ReferenceType::Author)?,
            author_id,
            *event_id,
        );
        buffer.put_cf(
            &self.reference_to_cf_handle(ReferenceType::Kind)?,
            kind_id,
            *event_id,
        );

        for tag in event.tags().iter() {
            match tag {
                Tag::PubKey(p) => {
                    let foreign_id = secondary_index.new_id(&p.id);
                    let local_id = secondary_index.new_id(&event_id);

                    buffer.put_cf(
                        &self.reference_to_cf_handle(ReferenceType::RefPublicKey)?,
                        foreign_id,
                        *event_id,
                    );
                    buffer.put_cf(
                        &self.reference_to_cf_handle(ReferenceType::RefEvent)?,
                        local_id,
                        &*p.id,
                    );
                }
                Tag::Event(e) => {
                    let foreign_id = secondary_index.new_id(&e.id);
                    let local_id = secondary_index.new_id(&event_id);

                    buffer.put_cf(
                        &self.reference_to_cf_handle(ReferenceType::RefEvent)?,
                        foreign_id,
                        *event_id,
                    );
                    buffer.put_cf(
                        &self.reference_to_cf_handle(ReferenceType::RefEvent)?,
                        local_id,
                        &*e.id,
                    );
                }
                _ => {}
            }
        }

        self.db.write(buffer).map_err(Error::RocksDb)?;

        Ok(true)
    }

    fn get_event<T: AsRef<[u8]>>(&self, id: T) -> Result<Option<Event>, Error> {
        Ok(self
            .db
            .get_cf(&self.reference_to_cf_handle(ReferenceType::Events)?, id)?
            .map(|event| serde_json::from_slice(&event))
            .transpose()?)
    }

    fn get_by_filter<C>(&self, mut query: Filter, for_each: C) -> Result<(), Error>
    where
        C: Fn(Event) -> Result<(), Error>,
    {
        let mut event_ids = vec![];
        let limit = if query.limit == 0 {
            None
        } else {
            Some(query.limit.try_into()?)
        };

        if !query.references_to_event.is_empty() {
            let ns = self.reference_to_cf_handle(ReferenceType::RefEvent)?;
            for public_key in query.references_to_event.iter() {
                event_ids.extend(self.get_event_referenced_as(&ns, public_key.as_ref(), limit)?);
            }
            query.references_to_event.clear();
        } else if !query.references_to_public_key.is_empty() {
            let ns = self.reference_to_cf_handle(ReferenceType::RefEvent)?;
            for public_key in query.references_to_public_key.iter() {
                event_ids.extend(self.get_event_referenced_as(&ns, public_key.as_ref(), limit)?);
            }
            query.references_to_public_key.clear();
        } else if !query.ids.is_empty() {
            query
                .ids
                .iter()
                .map(|id| event_ids.push(id.as_ref().to_vec()))
                .for_each(drop);
            query.ids.clear();
        } else if !query.authors.is_empty() {
            let ns = self.reference_to_cf_handle(ReferenceType::Author)?;
            for public_key in query.authors.iter() {
                event_ids.extend(self.get_event_ids_by_ns_and_prefix(
                    &ns,
                    public_key.as_ref(),
                    limit,
                )?);
            }
            query.authors.clear();
        } else if !query.kinds.is_empty() {
            let ns = self.reference_to_cf_handle(ReferenceType::Kind)?;
            for kind in query.kinds.iter() {
                let kind: u32 = (*kind).into();
                event_ids.extend(self.get_event_ids_by_ns_and_prefix(
                    &ns,
                    &kind.to_be_bytes(),
                    limit,
                )?);
            }
            query.kinds.clear();
        }

        load_events_and_filter(self, query, event_ids, for_each)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::{test, util::unique_nanoseconds};

    #[test]
    fn store_and_get() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::store_and_get(&db)
    }

    #[test]
    fn records_are_sorted_by_date_desc() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::records_are_sorted_by_date_desc(&db)
    }

    #[test]
    fn filter_by_references_zero_match() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::filter_by_references_zero_match(&db)
    }

    #[test]
    fn filter_by_references_and_kind() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::filter_by_references_and_kind(&db)
    }

    #[test]
    fn get_event_and_related_events() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::get_event_and_related_events(&db)
    }

    #[test]
    fn filter_by_author() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::filter_by_author(&db)
    }

    #[test]
    fn filter_by_authors() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::filter_by_authors(&db)
    }

    #[test]
    fn filter_kind() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::filter_kind(&db)
    }

    #[test]
    fn filter_by_references() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::filter_by_references(&db)
    }

    #[test]
    fn filter_by_author_and_kinds() {
        let db = RocksDb::new(format!("tests/db/{}", unique_nanoseconds())).expect("db");
        test::filter_by_author_and_kinds(&db)
    }
}
