use crate::{Error, Storage};
use chrono::{DateTime, Utc};
use nostr_rs_types::types::{Event, Filter, Id, Tag};
use rand::Rng;
use std::collections::HashSet;

/// Creates a secondary index for a given event ID. The secondary index is
/// sorted by created time in a descendent order, meaning that newer events will
/// be first.
///
/// This is useful to leverage natural sorting of the database, and to be able
/// to create secondary indexes to find by author, kind, tag, etc, and have that
/// result sorted by date naturally
pub struct SecondaryIndex<'a> {
    entry_id: &'a Id,
    sort_by: [u8; 8],
}

impl<'a> SecondaryIndex<'a> {
    pub fn new(entry_id: &'a Id, date_time: DateTime<Utc>) -> Self {
        // We need to reverse the number, to sort it. Otherwise, the first is
        // going to be the oldest entry, instead of the newest.
        let reversed_order_by = u64::MAX
            .checked_sub(date_time.timestamp() as u64)
            .unwrap_or_default()
            .to_be_bytes();
        Self {
            entry_id,
            sort_by: reversed_order_by,
        }
    }

    pub fn new_id<T: AsRef<[u8]>>(&self, index_by: T) -> Vec<u8> {
        let mut new_id = index_by.as_ref().to_vec();
        new_id.extend_from_slice(&self.sort_by);
        new_id.extend_from_slice(self.entry_id.as_ref());
        new_id
    }
}

// Get current nanoseconds and use the last 3 digits as a random number (because
// sometimes it comes as 0)
#[inline]
#[allow(dead_code)]
pub fn unique_nanoseconds() -> u128 {
    let mut rng = rand::thread_rng();
    let random_number = rng.gen_range(0..999);

    let ts = std::time::SystemTime::now()
        .duration_since(std::time::UNIX_EPOCH)
        .expect("time")
        .as_nanos();

    ts.checked_add(random_number).unwrap_or(ts)
}

/// Receives a list of event-IDs, and returns a list of events that match
/// the given filter.
///
/// The list of event-IDs should come from an index from any namespace, and
/// the filter will load the events from the `Events` namespace and will
/// filter them by the given parameters.
#[inline]
pub fn load_events_and_filter<T: Storage, C>(
    db: &T,
    mut query: Filter,
    mut event_ids: Vec<Vec<u8>>,
    for_each: C,
) -> Result<(), Error>
where
    C: Fn(Event) -> Result<(), Error>,
{
    if !query.ids.is_empty() {
        let mut ids = query.ids.iter().map(|id| id.as_ref()).collect::<Vec<_>>();
        ids.sort();
        event_ids.retain(|id| ids.binary_search(&id.as_slice()).is_ok());
    }

    let mut authors = query
        .authors
        .iter()
        .map(|id| id.as_ref())
        .collect::<Vec<_>>();

    let mut tag_primary_keys = query
        .references_to_public_key
        .iter()
        .map(|tag| tag.as_ref())
        .collect::<HashSet<_>>()
        .into_iter()
        .collect::<Vec<_>>();

    let mut tag_events = query
        .references_to_event
        .iter()
        .map(|tag| tag.as_ref())
        .collect::<HashSet<_>>()
        .into_iter()
        .collect::<Vec<_>>();

    // Sort everything for a quick binary_search instead of a linear search
    tag_events.sort();
    tag_primary_keys.sort();
    authors.sort();
    query.kinds.sort();

    let mut found = 0;

    for id in event_ids.iter() {
        if let Some(event) = db.get_event(id)? {
            if !tag_primary_keys.is_empty() || !tag_events.is_empty() && authors.is_empty() {
                let mut found = false;

                for tag in event.tags().iter() {
                    match tag {
                        Tag::Event(event) => {
                            if tag_events.binary_search(&event.id.as_ref()).is_ok() {
                                found = true;
                            }
                        }
                        Tag::PubKey(key) => {
                            if tag_primary_keys.binary_search(&key.id.as_ref()).is_ok() {
                                found = true;
                            }
                        }
                        _ => {}
                    }

                    if found {
                        break;
                    }
                }

                if !found {
                    continue;
                }
            }

            if !authors.is_empty() && authors.binary_search(&event.author().as_ref()).is_err() {
                continue;
            }
            if !query.kinds.is_empty() && query.kinds.binary_search(&event.kind()).is_err() {
                continue;
            }
            if let Some(since) = query.since {
                if event.created_at() < since {
                    continue;
                }
            }
            if let Some(until) = query.until {
                if event.created_at() > until {
                    continue;
                }
            }
            for_each(event)?;
            found += 1;
            if found == query.limit {
                break;
            }
        }
    }

    Ok(())
}
